@@ -1,9 +1,16 @@
+from __future__ import unicode_literals
+
 import logging
 import random
+import sys
 import threading
-import urllib
-import urllib2
-import urlparse
+try:                                                        # cover: disable
+    from urllib.parse import quote, urljoin, urlsplit, urlunsplit
+    from urllib.request import HTTPError, Request, URLError, urlopen
+except ImportError:
+    from urllib import quote
+    from urllib2 import HTTPError, Request, URLError, urlopen
+    from urlparse import urljoin, urlsplit, urlunsplit
 
 from django.conf import settings
 from django.core.files.base import ContentFile
@@ -16,7 +23,7 @@ from .settings import get_setting
 logger = logging.getLogger(__name__)
 
 
-class UnexpectedStatusCode(urllib2.HTTPError):
+class UnexpectedStatusCode(HTTPError):
 
     """Exception raised when a server returns an unexpected status code.
 
@@ -34,26 +41,26 @@ class UnexpectedStatusCode(urllib2.HTTPError):
             resp.url, resp.code, resp.msg, resp.headers, resp.fp)
 
 
-class GetRequest(urllib2.Request):
+class GetRequest(Request):
     """HTTP GET request."""
-    # This adds nothing to urllib2, but it's there for consistency.
+    # This adds nothing to urllib, but it's there for consistency.
     def get_method(self):
         return 'GET'
 
 
-class HeadRequest(urllib2.Request):
+class HeadRequest(Request):
     """HTTP HEAD request."""
     def get_method(self):
         return 'HEAD'
 
 
-class DeleteRequest(urllib2.Request):
+class DeleteRequest(Request):
     """HTTP DELETE request."""
     def get_method(self):
         return 'DELETE'
 
 
-class PutRequest(urllib2.Request):
+class PutRequest(Request):
     """HTTP PUT request."""
     def get_method(self):
         return 'PUT'
@@ -69,24 +76,32 @@ class DefaultTransport(object):
     timeout = get_setting('TIMEOUT')
 
     def __init__(self, base_url):
-        scheme, netloc, path, query, fragment = urlparse.urlsplit(base_url)
+        scheme, netloc, path, query, fragment = urlsplit(base_url)
         if query or fragment:
             raise ValueError('base_url may not contain a query or fragment.')
         self.scheme = scheme or 'http'
         self.path = path or '/'
 
-    def get_url(self, host, name):
-        """Return the full URL for a file on a given host (internal use)."""
-        path = self.path + urllib.quote(name.encode('utf-8'))
-        return urlparse.urlunsplit((self.scheme, host, path, '', ''))
+    ### Hooks for custom transports
+
+    def _get_url(self, host, name):
+        """Return the full URL for a file on a given host."""
+        path = self.path + quote(name.encode('utf-8'))
+        return urlunsplit((self.scheme, host, path, '', ''))
+
+    def _http_request(self, request):
+        """Return a response object for a given request."""
+        return urlopen(request, timeout=self.timeout)
+
+    ### Wrappers around HTTP methods
 
     def content(self, host, name):
         """Get the content of a file as a string.
 
-        urllib2.URLError will be raised if something goes wrong.
+        URLError will be raised if something goes wrong.
         """
-        url = self.get_url(host, name)
-        resp = urllib2.urlopen(GetRequest(url), timeout=self.timeout)
+        url = self._get_url(host, name)
+        resp = self._http_request(GetRequest(url))
         if resp.code != 200:
             raise UnexpectedStatusCode(resp)
         length = resp.info().get('Content-Length')
@@ -100,15 +115,15 @@ class DefaultTransport(object):
 
         Return True if the file exists, False if it doesn't.
 
-        urllib2.URLError will be raised if something goes wrong.
+        URLError will be raised if something goes wrong.
         """
-        url = self.get_url(host, name)
+        url = self._get_url(host, name)
         try:
-            resp = urllib2.urlopen(HeadRequest(url), timeout=self.timeout)
+            resp = self._http_request(HeadRequest(url))
             if resp.code != 200:
                 raise UnexpectedStatusCode(resp)
             return True
-        except urllib2.HTTPError, e:
+        except HTTPError as e:
             if e.code not in (404, 410):
                 raise
             return False
@@ -118,10 +133,10 @@ class DefaultTransport(object):
 
         This method relies on the Content-Length header.
 
-        urllib2.URLError will be raised if something goes wrong.
+        URLError will be raised if something goes wrong.
         """
-        url = self.get_url(host, name)
-        resp = urllib2.urlopen(HeadRequest(url), timeout=self.timeout)
+        url = self._get_url(host, name)
+        resp = self._http_request(HeadRequest(url))
         if resp.code != 200:
             raise UnexpectedStatusCode(resp)
         length = resp.info().get('Content-Length')
@@ -135,10 +150,10 @@ class DefaultTransport(object):
 
         Return True if the file existed, False if it did not.
 
-        urllib2.URLError will be raised if something goes wrong.
+        URLError will be raised if something goes wrong.
         """
-        url = self.get_url(host, name)
-        resp = urllib2.urlopen(PutRequest(url, content), timeout=self.timeout)
+        url = self._get_url(host, name)
+        resp = self._http_request(PutRequest(url, content))
         if resp.code == 201:
             return False
         elif resp.code == 204:
@@ -152,15 +167,15 @@ class DefaultTransport(object):
 
         Return True if the file existed, False if it did not.
 
-        urllib2.URLError will be raised if something goes wrong.
+        URLError will be raised if something goes wrong.
         """
-        url = self.get_url(host, name)
+        url = self._get_url(host, name)
         try:
-            resp = urllib2.urlopen(DeleteRequest(url), timeout=self.timeout)
+            resp = self._http_request(DeleteRequest(url))
             if resp.code not in (200, 204):
                 raise UnexpectedStatusCode(resp)
             return True
-        except urllib2.HTTPError, e:
+        except HTTPError as e:
             if e.code not in (404, 410):
                 raise
             logger.warning("DELETE on missing file %s on %s.", name, host)
@@ -172,95 +187,45 @@ class DistributedStorageMixin(object):
     """Mixin for storage backends that distribute files on several servers."""
 
     fatal_exceptions = get_setting('FATAL_EXCEPTIONS')
-    exc_info = get_setting('SHOW_TRACEBACK')
+    show_traceback = get_setting('SHOW_TRACEBACK')
 
-    def __init__(self, hosts=None, base_url=None):
+    def __init__(self, hosts=None, base_url=None, transport=DefaultTransport):
         if hosts is None:                                   # cover: disable
             hosts = get_setting('MEDIA_HOSTS')
         self.hosts = hosts
         if base_url is None:                                # cover: disable
             base_url = settings.MEDIA_URL
         self.base_url = base_url
-        self.transport = DefaultTransport(base_url=self.base_url)
+        self.transport = transport(base_url=base_url)
 
-    def execute_parallel(self, func, url, *extra):
+    def execute(self, func, url, *args, **kwargs):
         """Run an action over several hosts in parallel."""
         exceptions = {}
 
-        def execute_one(host):
+        def execute_inner(host):
             try:
-                func(host, url, *extra)
-            except Exception, exc:
-                exceptions[host] = exc
+                func(host, url, *args, **kwargs)
+            except Exception:
+                exceptions[host] = sys.exc_info()
 
-        threads = [threading.Thread(target=execute_one, args=(host,))
-                for host in self.hosts]
-        for thread in threads:
-            thread.start()
-        for thread in threads:
-            thread.join()
+        if len(self.hosts) == 1:
+            execute_inner(self.hosts[0])
+        else:
+            threads = [threading.Thread(target=execute_inner, args=(host,))
+                    for host in self.hosts]
+            for thread in threads:
+                thread.start()
+            for thread in threads:
+                thread.join()
 
-        for host, exception in exceptions.iteritems():
-            logger.error("Failed to %s %s on %s.", func.func_name, url, host,
-                exc_info=self.exc_info)
+        for host, exc_info in exceptions.items():
+            action = func.__name__
+            logger.error("Failed to %s %s on %s.", action, url, host,
+                    exc_info=exc_info if self.show_traceback else None)
 
         if exceptions and self.fatal_exceptions:
-            # Let's raise the first exception, we've logged them all anyway
-            raise exceptions.itervalues().next()
-
-    ### Hooks for custom storage objects
-
-    def _open(self, name, mode='rb'):
-        # Allowing writes would be doable, if we distribute the file to the
-        # media servers when it's closed. Let's forbid it for now.
-        if mode != 'rb':
-            raise IOError('Unsupported mode %r, use %r.' % (mode, 'rb'))
-        host = random.choice(self.hosts)
-        try:
-            return ContentFile(self.transport.content(host, name))
-        except urllib2.URLError:
-            logger.error("Failed to download %s from %s.", name, host,
-                    exc_info=self.exc_info)
-            raise
-
-    def _save(self, name, content):
-        # It's hard to avoid buffering the whole file in memory,
-        # because different threads will read it simultaneously.
-        self.execute_parallel(self.transport.create, name, content.read())
-        return name
-
-    # The implementations of get_valid_name, get_available_name, path and url
-    # in Storage and FileSystemStorage are OK for DistributedStorage and
-    # HybridStorage.
-
-    ### Mandatory methods
-
-    def delete(self, name):
-        self.execute_parallel(self.transport.delete, name)
-
-    def exists(self, name):
-        host = random.choice(self.hosts)
-        try:
-            return self.transport.exists(host, name)
-        except urllib2.URLError:
-            logger.error("Failed to check if %s exists on %s.", name, host,
-                    exc_info=self.exc_info)
-            raise
-
-    # It is not possible to implement listdir in pure HTTP. It could
-    # be done with WebDAV.
-
-    def size(self, name):
-        host = random.choice(self.hosts)
-        try:
-            return self.transport.size(host, name)
-        except urllib2.URLError:
-            logger.error("Failed to get the size of %s from %s.", name, host,
-                    exc_info=self.exc_info)
-            raise
-
-    def url(self, name):
-        return urlparse.urljoin(self.base_url, filepath_to_uri(name))
+            # Let's raise a random exception, we've logged them all anyway
+            raise exceptions.popitem()[1][1]
 
 
 class DistributedStorage(DistributedStorageMixin, Storage):
@@ -277,6 +242,59 @@ class DistributedStorage(DistributedStorageMixin, Storage):
             logger.warning("You have been warned.")
         Storage.__init__(self)
 
+    ### Hooks for custom storage objects
+
+    def _open(self, name, mode='rb'):
+        # Allowing writes would be doable, if we distribute the file to the
+        # media servers when it's closed. Let's forbid it for now.
+        if mode != 'rb':                                    # cover: disable
+            raise IOError('Unsupported mode %r, use %r.' % (mode, 'rb'))
+        host = random.choice(self.hosts)
+        try:
+            return ContentFile(self.transport.content(host, name))
+        except URLError:
+            logger.error("Failed to download %s from %s.", name, host,
+                    exc_info=self.show_traceback)
+            raise
+
+    def _save(self, name, content):
+        # It's hard to avoid buffering the whole file in memory,
+        # because different threads will read it simultaneously.
+        self.execute(self.transport.create, name, content.read())
+        return name
+
+    ### Mandatory methods
+
+    # The implementations of get_valid_name, get_available_name, and path
+    # in Storage are OK for DistributedStorage.
+
+    def delete(self, name):
+        self.execute(self.transport.delete, name)
+
+    def exists(self, name):
+        host = random.choice(self.hosts)
+        try:
+            return self.transport.exists(host, name)
+        except URLError:
+            logger.error("Failed to check if %s exists on %s.", name, host,
+                    exc_info=self.show_traceback)
+            raise
+
+    # It is not possible to implement listdir in pure HTTP. It could
+    # be done with WebDAV.
+
+    def size(self, name):
+        host = random.choice(self.hosts)
+        try:
+            return self.transport.size(host, name)
+        except URLError:
+            logger.error("Failed to get the size of %s from %s.", name, host,
+                    exc_info=self.show_traceback)
+            raise
+
+    def url(self, name):
+        return urljoin(self.base_url, filepath_to_uri(name))
+
 
 class HybridStorage(DistributedStorageMixin, FileSystemStorage):
 
@@ -293,31 +311,50 @@ class HybridStorage(DistributedStorageMixin, FileSystemStorage):
     ### Hooks for custom storage objects
 
     def _open(self, name, mode='rb'):
-        # Writing is forbidden, see DistributedStorageMixin._open.
-        if mode != 'rb':
+        # Writing is forbidden, see DistributedStorage._open.
+        if mode != 'rb':                                    # cover: disable
             raise IOError('Unsupported mode %r, use %r.' % (mode, 'rb'))
         return FileSystemStorage._open(self, name, mode)
 
     def _save(self, name, content):
         name = FileSystemStorage._save(self, name, content)
-        content.seek(0)
         # After this line, we will assume that 'name' is available on the
         # media servers. This could be wrong if a delete for this file name
         # failed at some point in the past.
-        DistributedStorageMixin._save(self, name, content)
+        self.execute(self.upload, name)
         return name
 
+    # Make this a separate method so it can be passed to a task queue.
+    def upload(self, host, name):
+        with self.open(name) as handle:
+            self.transport.create(host, name, handle.read())
+
     ### Mandatory methods
 
+    # The implementations of get_valid_name, get_available_name, path, exists,
+    # listdir, size, and url in FileSystemStorage are OK for HybridStorage.
+
     def delete(self, name):
         FileSystemStorage.delete(self, name)
-        DistributedStorageMixin.delete(self, name)
+        self.execute(self.transport.delete, name)
 
-    def exists(self, name):
-        return FileSystemStorage.exists(self, name)
 
-    def listdir(self, name):
-        return FileSystemStorage.listdir(self, name)
+class AsyncStorage(HybridStorage):
 
-    def size(self, name):
-        return FileSystemStorage.size(self, name)
+    """Backend that stores files both locally and remotely over HTTP."""
+
+    def execute(self, func, url, *args, **kwargs):
+        """Run an action over several hosts asynchronously."""
+        for host in self.hosts:
+            self.execute_one(func, host, url, *args, **kwargs)
+
+    def execute_one(self, func, host, url, *args, **kwargs):
+        """Run a single action asynchronously."""
+        def execute_inner():
+            try:
+                func(host, url, *args, **kwargs)
+            except Exception:
+                action = func.__name__
+                logger.error("Failed to %s %s on %s.", action, url, host,
+                        exc_info=self.show_traceback)
+        return threading.Thread(target=execute_inner).start()
